﻿// Microsoft Public License (Ms-PL)
// See the file License.rtf or License.txt for the license details.

// Copyright (c) 2011, Semyon A. Chertkov (semyonc@gmail.com)
// All rights reserved.

using System;
using System.Collections.Generic;
using System.Xml;
using System.Xml.XPath;
using System.Xml.Schema;

using Wmhelp.XPath2.Value;
using Wmhelp.XPath2.MS;

namespace Wmhelp.XPath2
{
    internal delegate object XPathFunctionDelegate(XPath2Context context, IContextProvider provider, object[] args);

    public class FunctionTable
    {
        private Dictionary<FunctionDesc, XPathFunctionDelegate> _funcTable;

        private FunctionTable()
        {
            _funcTable = new Dictionary<FunctionDesc, XPathFunctionDelegate>();
            Add(XmlReservedNs.NsXQueryFunc, "dateTime", 2, (context, provider, args) => 
                ExtFuncs.CreateDateTime(
                   CoreFuncs.CastArg(context, args[0], new SequenceType(XmlTypeCode.Date, XmlTypeCardinality.ZeroOrOne)),
                   CoreFuncs.CastArg(context, args[1], new SequenceType(XmlTypeCode.Time, XmlTypeCardinality.ZeroOrOne))));
            Add(XmlReservedNs.NsXQueryFunc, "current-dateTime", 0, (context, provider, args) => 
                    ExtFuncs.GetCurrentDateTime(context));
            Add(XmlReservedNs.NsXQueryFunc, "current-date", 0, (context, provider, args) => 
                    ExtFuncs.GetCurrentDate(context));
            Add(XmlReservedNs.NsXQueryFunc, "current-time", 0, (context, provider, args) => 
                    ExtFuncs.GetCurrentTime(context));
            Add(XmlReservedNs.NsXQueryFunc, "in-scope-prefixes", 1, (context, provider, args) => 
                ExtFuncs.GetInScopePrefixes(CoreFuncs.NodeValue(args[0])));
            Add(XmlReservedNs.NsXQueryFunc, "namespace-uri-for-prefix", 2, (context, provider, args) => 
                ExtFuncs.GetNamespaceUriForPrefix(context, CoreFuncs.Atomize(args[0]), CoreFuncs.NodeValue(args[1])));
            Add(XmlReservedNs.NsXQueryFunc, "resolve-QName", 2, (context, provider, args) =>
                ExtFuncs.ResolveQName(context, CoreFuncs.Atomize(args[0]), CoreFuncs.NodeValue(args[1])));
            Add(XmlReservedNs.NsXQueryFunc, "QName", 2, (context, provider, args) =>
                ExtFuncs.CreateQName(context, CoreFuncs.CastString(context, args[0]),
                    (String)CoreFuncs.CastArg(context, args[1], new SequenceType(XmlTypeCode.String))));
            Add(XmlReservedNs.NsXQueryFunc, "prefix-from-QName", 1, (context, provider, args) =>
                ExtFuncs.PrefixFromQName(CoreFuncs.Atomize(args[0])));
            Add(XmlReservedNs.NsXQueryFunc, "local-name-from-QName", 1, (context, provider, args) =>
                ExtFuncs.LocalNameFromQName(CoreFuncs.Atomize(args[0])));
            Add(XmlReservedNs.NsXQueryFunc, "namespace-uri-from-QName", 1, (context, provider, args) => 
                ExtFuncs.NamespaceUriFromQName(CoreFuncs.Atomize(args[0])));
            Add(XmlReservedNs.NsXQueryFunc, "string-to-codepoints", 1, (context, provider, args) => 
                ExtFuncs.StringToCodepoint(CoreFuncs.CastString(context, args[0])));
            Add(XmlReservedNs.NsXQueryFunc, "codepoints-to-string", 1, (context, provider, args) => 
                ExtFuncs.CodepointToString((XPath2NodeIterator)CoreFuncs.CastArg(context, args[0],  
                    new SequenceType(XmlTypeCode.Int, XmlTypeCardinality.ZeroOrMore))));
            Add(XmlReservedNs.NsXQueryFunc, "default-collation", 0, (context, provider, args) => 
                ExtFuncs.DefaultCollation(context));
            Add(XmlReservedNs.NsXQueryFunc, "resolve-uri", 2, (context, provider, args) => 
                ExtFuncs.ResolveUri(CoreFuncs.CastString(context, args[0]),
                    CoreFuncs.CastString(context, args[1])));
            Add(XmlReservedNs.NsXQueryFunc, "resolve-uri", 1, (context, provider, args) => 
                ExtFuncs.ResolveUri(context, CoreFuncs.CastString(context, args[0])));
            Add(XmlReservedNs.NsXQueryFunc, "static-base-uri", 0, (context, provider, args) => 
                ExtFuncs.StaticBaseUri(context));
            Add(XmlReservedNs.NsXQueryFunc, "implicit-timezone", 0, (context, provider, args) => 
                ExtFuncs.ImplicitTimezone(context));
            Add(XmlReservedNs.NsXQueryFunc, "lang", 2, (context, provider, args) => 
                ExtFuncs.NodeLang(CoreFuncs.CastString(context, args[0]), CoreFuncs.NodeValue(args[1])));
            Add(XmlReservedNs.NsXQueryFunc, "lang", 1, (context, provider, args) =>
                ExtFuncs.NodeLang(provider, CoreFuncs.CastString(context, args[0])));
            Add(XmlReservedNs.NsXQueryFunc, "name", 1, (context, provider, args) => 
                ExtFuncs.GetName(CoreFuncs.NodeValue(args[0], false)));
            Add(XmlReservedNs.NsXQueryFunc, "name", 0, (context, provider, args) => 
                ExtFuncs.GetName(provider));
            Add(XmlReservedNs.NsXQueryFunc, "node-name", 1, (context, provider, args) => 
                ExtFuncs.GetNodeName(context, CoreFuncs.NodeValue(args[0], false)));
            Add(XmlReservedNs.NsXQueryFunc, "local-name", 1, (context, provider, args) => 
                ExtFuncs.GetLocalName(CoreFuncs.NodeValue(args[0], false)));
            Add(XmlReservedNs.NsXQueryFunc, "local-name", 0, (context, provider, args) => 
                ExtFuncs.GetLocalName(provider));
            Add(XmlReservedNs.NsXQueryFunc, "namespace-uri", 1, (context, provider, args) => 
                ExtFuncs.GetNamespaceUri(CoreFuncs.NodeValue(args[0], false)));
            Add(XmlReservedNs.NsXQueryFunc, "namespace-uri", 0, (context, provider, args) => 
                ExtFuncs.GetNamespaceUri(provider));
            Add(XmlReservedNs.NsXQueryFunc, "nilled", 1, (context, provider, args) => 
                ExtFuncs.GetNilled(CoreFuncs.NodeValue(args[0], false)));
            Add(XmlReservedNs.NsXQueryFunc, "base-uri", 0, (context, provider, args) => 
                ExtFuncs.GetBaseUri(context, provider));
            Add(XmlReservedNs.NsXQueryFunc, "base-uri", 1, (context, provider, args) => 
                ExtFuncs.GetBaseUri(context, CoreFuncs.NodeValue(args[0], false)));
            Add(XmlReservedNs.NsXQueryFunc, "document-uri", 1, (context, provider, args) => 
                ExtFuncs.DocumentUri(CoreFuncs.NodeValue(args[0], false)));
            Add(XmlReservedNs.NsXQueryFunc, "trace", 1, (context, provider, args) => 
                ExtFuncs.WriteTrace(context, XPath2NodeIterator.Create(args[0])));
            Add(XmlReservedNs.NsXQueryFunc, "trace", 2, (context, provider, args) => 
                ExtFuncs.WriteTrace(context, XPath2NodeIterator.Create(args[0]),
                (String)CoreFuncs.CastArg(context, args[1], new SequenceType(XmlTypeCode.String))));
            Add(XmlReservedNs.NsXQueryFunc, "data", 1, (context, provider, args) => 
                ExtFuncs.GetData(XPath2NodeIterator.Create(args[0])));
            Add(XmlReservedNs.NsXQueryFunc, "concat", (context, provider, args) => 
                ExtFuncs.Concat(context, args));
            Add(XmlReservedNs.NsXQueryFunc, "string-join", 2, (context, provider, args) =>
                ExtFuncs.StringJoin(XPath2NodeIterator.Create(args[0]),
                   (String)CoreFuncs.CastArg(context, args[1], new SequenceType(XmlTypeCode.String))));
            Add(XmlReservedNs.NsXQueryFunc, "substring", 3, (context, provider, args) =>
                ExtFuncs.Substring(CoreFuncs.CastString(context, args[0]), 
                    CoreFuncs.Number(context, args[1]), CoreFuncs.Number(context, args[2])));
            Add(XmlReservedNs.NsXQueryFunc, "substring", 2, (context, provider, args) =>
                ExtFuncs.Substring(CoreFuncs.CastString(context, args[0]), 
                    CoreFuncs.Number(context, args[1])));
            Add(XmlReservedNs.NsXQueryFunc, "string-length", 0, (context, provider, args) => 
                ExtFuncs.StringLength(context, provider));
            Add(XmlReservedNs.NsXQueryFunc, "string-length", 1, (context, provider, args) =>
                ExtFuncs.StringLength(CoreFuncs.CastString(context, args[0])));
            Add(XmlReservedNs.NsXQueryFunc, "normalize-space", 0, (context, provider, args) => 
                ExtFuncs.NormalizeSpace(context, provider));
            Add(XmlReservedNs.NsXQueryFunc, "normalize-space", 1, (context, provider, args) =>
                CoreFuncs.NormalizeSpace(CoreFuncs.CastString(context, args[0])));
            Add(XmlReservedNs.NsXQueryFunc, "normalize-unicode", 1, (context, provider, args) =>
                ExtFuncs.NormalizeUnicode(CoreFuncs.CastString(context, args[0])));
            Add(XmlReservedNs.NsXQueryFunc, "normalize-unicode", 2, (context, provider, args) =>
                ExtFuncs.NormalizeUnicode(CoreFuncs.CastString(context, args[0]),
                    (String)CoreFuncs.CastArg(context, CoreFuncs.Atomize(args[1]), new SequenceType(XmlTypeCode.String))));
            Add(XmlReservedNs.NsXQueryFunc, "upper-case", 1, (context, provider, args) =>
                ExtFuncs.UpperCase(CoreFuncs.CastString(context, args[0])));
            Add(XmlReservedNs.NsXQueryFunc, "lower-case", 1, (context, provider, args) =>
                ExtFuncs.LowerCase(CoreFuncs.CastString(context, args[0])));
            Add(XmlReservedNs.NsXQueryFunc, "translate", 3, (context, provider, args) =>
                ExtFuncs.Translate(CoreFuncs.CastString(context, args[0]),
                    (String)CoreFuncs.CastArg(context, CoreFuncs.Atomize(args[1]), new SequenceType(XmlTypeCode.String)),
                    (String)CoreFuncs.CastArg(context, CoreFuncs.Atomize(args[2]), new SequenceType(XmlTypeCode.String))));
            Add(XmlReservedNs.NsXQueryFunc, "encode-for-uri", 1, (context, provider, args) =>
                ExtFuncs.EncodeForUri(CoreFuncs.CastString(context, args[0])));
            Add(XmlReservedNs.NsXQueryFunc, "iri-to-uri", 1, (context, provider, args) =>
                ExtFuncs.IriToUri(CoreFuncs.CastString(context, args[0])));
            Add(XmlReservedNs.NsXQueryFunc, "escape-html-uri", 1, (context, provider, args) =>
                ExtFuncs.EscapeHtmlUri(CoreFuncs.CastString(context, args[0])));
            Add(XmlReservedNs.NsXQueryFunc, "contains", 3, (context, provider, args) =>
                ExtFuncs.Contains(context, CoreFuncs.CastString(context, args[0]),
                    CoreFuncs.CastString(context, args[1]), 
                    (String)CoreFuncs.CastArg(context, args[2], new SequenceType(XmlTypeCode.String))));
            Add(XmlReservedNs.NsXQueryFunc, "contains", 2, (context, provider, args) =>
                ExtFuncs.Contains(CoreFuncs.CastString(context, args[0]),
                    CoreFuncs.CastString(context, args[1])));
            Add(XmlReservedNs.NsXQueryFunc, "starts-with", 3, (context, provider, args) =>
                ExtFuncs.StartsWith(context, CoreFuncs.CastString(context, args[0]),
                    CoreFuncs.CastString(context, args[1]),
                    (String)CoreFuncs.CastArg(context, CoreFuncs.Atomize(args[2]), new SequenceType(XmlTypeCode.String))));
            Add(XmlReservedNs.NsXQueryFunc, "starts-with", 2, (context, provider, args) =>
                ExtFuncs.StartsWith(CoreFuncs.CastString(context, args[0]),
                    CoreFuncs.CastString(context, args[1])));
            Add(XmlReservedNs.NsXQueryFunc, "ends-with", 3, (context, provider, args) =>
                ExtFuncs.EndsWith(context, CoreFuncs.CastString(context, args[0]),
                    CoreFuncs.CastString(context, args[1]),
                    (String)CoreFuncs.CastArg(context, CoreFuncs.Atomize(args[2]), new SequenceType(XmlTypeCode.String))));
            Add(XmlReservedNs.NsXQueryFunc, "ends-with", 2, (context, provider, args) =>
                ExtFuncs.EndsWith(CoreFuncs.CastString(context, args[0]),
                    CoreFuncs.CastString(context, args[1])));
            Add(XmlReservedNs.NsXQueryFunc, "substring-before", 3, (context, provider, args) =>
                ExtFuncs.SubstringBefore(context, CoreFuncs.CastString(context, args[0]),
                    CoreFuncs.CastString(context, args[1]),
                    (String)CoreFuncs.CastArg(context, CoreFuncs.Atomize(args[2]), new SequenceType(XmlTypeCode.String))));
            Add(XmlReservedNs.NsXQueryFunc, "substring-before", 2, (context, provider, args) =>
                ExtFuncs.SubstringBefore(CoreFuncs.CastString(context, args[0]),
                    CoreFuncs.CastString(context, args[1])));
            Add(XmlReservedNs.NsXQueryFunc, "substring-after", 3, (context, provider, args) =>
                ExtFuncs.SubstringAfter(context, CoreFuncs.CastString(context, args[0]),
                    CoreFuncs.CastString(context, args[1]),
                    (String)CoreFuncs.CastArg(context, CoreFuncs.Atomize(args[2]), new SequenceType(XmlTypeCode.String))));
            Add(XmlReservedNs.NsXQueryFunc, "substring-after", 2, (context, provider, args) =>
                ExtFuncs.SubstringAfter(CoreFuncs.CastString(context, args[0]),
                    CoreFuncs.CastString(context, args[1])));
            Add(XmlReservedNs.NsXQueryFunc, "matches", 2, (context, provider, args) =>
                ExtFuncs.Matches(CoreFuncs.CastString(context, args[0]),
                    CoreFuncs.CastString(context, args[1])));
            Add(XmlReservedNs.NsXQueryFunc, "matches", 3, (context, provider, args) =>
                ExtFuncs.Matches(CoreFuncs.CastString(context, args[0]),
                    CoreFuncs.CastString(context, args[1]),
                    (String)CoreFuncs.CastArg(context, CoreFuncs.Atomize(args[2]), new SequenceType(XmlTypeCode.String))));
            Add(XmlReservedNs.NsXQueryFunc, "replace", 3, (context, provider, args) =>
                ExtFuncs.Replace(CoreFuncs.CastString(context, args[0]),
                    CoreFuncs.CastString(context, args[1]),
                    (String)CoreFuncs.CastArg(context, CoreFuncs.Atomize(args[2]), new SequenceType(XmlTypeCode.String))));
            Add(XmlReservedNs.NsXQueryFunc, "replace", 4, (context, provider, args) => 
                ExtFuncs.Replace(CoreFuncs.CastString(context, args[0]),
                    CoreFuncs.CastString(context, args[1]),
                    (String)CoreFuncs.CastArg(context, CoreFuncs.Atomize(args[2]), new SequenceType(XmlTypeCode.String)),
                    (String)CoreFuncs.CastArg(context, CoreFuncs.Atomize(args[3]), new SequenceType(XmlTypeCode.String))));
            Add(XmlReservedNs.NsXQueryFunc, "tokenize", 2, (context, provider, args) => 
                ExtFuncs.Tokenize(CoreFuncs.CastString(context, args[0]),
                    CoreFuncs.CastString(context, args[1])));
            Add(XmlReservedNs.NsXQueryFunc, "tokenize", 3, (context, provider, args) => 
                ExtFuncs.Tokenize(CoreFuncs.CastString(context, args[0]),
                    CoreFuncs.CastString(context, args[1]),
                    (String)CoreFuncs.CastArg(context, CoreFuncs.Atomize(args[2]), new SequenceType(XmlTypeCode.String))));
            Add(XmlReservedNs.NsXQueryFunc, "years-from-duration", 1, (context, provider, args) => 
                ExtFuncs.YearsFromDuration(CoreFuncs.CastArg(context, CoreFuncs.Atomize(args[0]), 
                    new SequenceType(XmlTypeCode.Duration, XmlTypeCardinality.ZeroOrOne))));
            Add(XmlReservedNs.NsXQueryFunc, "months-from-duration", 1, (context, provider, args) =>
                ExtFuncs.MonthsFromDuration(CoreFuncs.CastArg(context, CoreFuncs.Atomize(args[0]),
                    new SequenceType(XmlTypeCode.Duration, XmlTypeCardinality.ZeroOrOne))));
            Add(XmlReservedNs.NsXQueryFunc, "days-from-duration", 1, (context, provider, args) =>
                ExtFuncs.DaysFromDuration(CoreFuncs.CastArg(context, CoreFuncs.Atomize(args[0]),
                    new SequenceType(XmlTypeCode.Duration, XmlTypeCardinality.ZeroOrOne))));
            Add(XmlReservedNs.NsXQueryFunc, "hours-from-duration", 1, (context, provider, args) =>
                ExtFuncs.HoursFromDuration(CoreFuncs.CastArg(context, CoreFuncs.Atomize(args[0]),
                    new SequenceType(XmlTypeCode.Duration, XmlTypeCardinality.ZeroOrOne))));
            Add(XmlReservedNs.NsXQueryFunc, "minutes-from-duration", 1, (context, provider, args) =>
                ExtFuncs.MinutesFromDuration(CoreFuncs.CastArg(context, CoreFuncs.Atomize(args[0]),
                    new SequenceType(XmlTypeCode.Duration, XmlTypeCardinality.ZeroOrOne))));
            Add(XmlReservedNs.NsXQueryFunc, "seconds-from-duration", 1, (context, provider, args) =>
                ExtFuncs.SecondsFromDuration(CoreFuncs.CastArg(context, CoreFuncs.Atomize(args[0]),
                    new SequenceType(XmlTypeCode.Duration, XmlTypeCardinality.ZeroOrOne))));
            Add(XmlReservedNs.NsXQueryFunc, "year-from-dateTime", 1, (context, provider, args) => 
                ExtFuncs.YearFromDateTime(CoreFuncs.CastArg(context, CoreFuncs.Atomize(args[0]),
                    new SequenceType(XmlTypeCode.DateTime, XmlTypeCardinality.ZeroOrOne))));
            Add(XmlReservedNs.NsXQueryFunc, "month-from-dateTime", 1, (context, provider, args) =>
                ExtFuncs.MonthFromDateTime(CoreFuncs.CastArg(context, CoreFuncs.Atomize(args[0]),
                    new SequenceType(XmlTypeCode.DateTime, XmlTypeCardinality.ZeroOrOne))));
            Add(XmlReservedNs.NsXQueryFunc, "day-from-dateTime", 1, (context, provider, args) =>
                ExtFuncs.DayFromDateTime(CoreFuncs.CastArg(context, CoreFuncs.Atomize(args[0]),
                    new SequenceType(XmlTypeCode.DateTime, XmlTypeCardinality.ZeroOrOne))));
            Add(XmlReservedNs.NsXQueryFunc, "hours-from-dateTime", 1, (context, provider, args) =>
                ExtFuncs.HoursFromDateTime(CoreFuncs.CastArg(context, CoreFuncs.Atomize(args[0]),
                    new SequenceType(XmlTypeCode.DateTime, XmlTypeCardinality.ZeroOrOne))));
            Add(XmlReservedNs.NsXQueryFunc, "minutes-from-dateTime", 1, (context, provider, args) =>
                ExtFuncs.MinutesFromDateTime(CoreFuncs.CastArg(context, CoreFuncs.Atomize(args[0]),
                    new SequenceType(XmlTypeCode.DateTime, XmlTypeCardinality.ZeroOrOne))));
            Add(XmlReservedNs.NsXQueryFunc, "seconds-from-dateTime", 1, (context, provider, args) =>
                ExtFuncs.SecondsFromDateTime(CoreFuncs.CastArg(context, CoreFuncs.Atomize(args[0]),
                    new SequenceType(XmlTypeCode.DateTime, XmlTypeCardinality.ZeroOrOne))));
            Add(XmlReservedNs.NsXQueryFunc, "timezone-from-dateTime", 1, (context, provider, args) =>
               ExtFuncs.TimezoneFromDateTime(CoreFuncs.CastArg(context, CoreFuncs.Atomize(args[0]),
                    new SequenceType(XmlTypeCode.DateTime, XmlTypeCardinality.ZeroOrOne))));
            Add(XmlReservedNs.NsXQueryFunc, "year-from-date", 1, (context, provider, args) =>
                ExtFuncs.YearFromDate(CoreFuncs.CastArg(context, CoreFuncs.Atomize(args[0]),
                    new SequenceType(XmlTypeCode.Date, XmlTypeCardinality.ZeroOrOne))));
            Add(XmlReservedNs.NsXQueryFunc, "month-from-date", 1, (context, provider, args) =>
                ExtFuncs.MonthFromDate(CoreFuncs.CastArg(context, CoreFuncs.Atomize(args[0]),
                    new SequenceType(XmlTypeCode.Date, XmlTypeCardinality.ZeroOrOne))));
            Add(XmlReservedNs.NsXQueryFunc, "day-from-date", 1, (context, provider, args) =>
                ExtFuncs.DayFromDate(CoreFuncs.CastArg(context, CoreFuncs.Atomize(args[0]),
                    new SequenceType(XmlTypeCode.Date, XmlTypeCardinality.ZeroOrOne))));
            Add(XmlReservedNs.NsXQueryFunc, "timezone-from-date", 1, (context, provider, args) =>
                ExtFuncs.TimezoneFromDate(CoreFuncs.CastArg(context, CoreFuncs.Atomize(args[0]),
                    new SequenceType(XmlTypeCode.Date, XmlTypeCardinality.ZeroOrOne))));
            Add(XmlReservedNs.NsXQueryFunc, "hours-from-time", 1, (context, provider, args) => 
                ExtFuncs.HoursFromTime(CoreFuncs.CastArg(context, CoreFuncs.Atomize(args[0]),
                    new SequenceType(XmlTypeCode.Time, XmlTypeCardinality.ZeroOrOne))));
            Add(XmlReservedNs.NsXQueryFunc, "minutes-from-time", 1, (context, provider, args) =>
                ExtFuncs.MinutesFromTime(CoreFuncs.CastArg(context, CoreFuncs.Atomize(args[0]),
                    new SequenceType(XmlTypeCode.Time, XmlTypeCardinality.ZeroOrOne))));
            Add(XmlReservedNs.NsXQueryFunc, "seconds-from-time", 1, (context, provider, args) =>
                ExtFuncs.SecondsFromTime(CoreFuncs.CastArg(context, CoreFuncs.Atomize(args[0]),
                    new SequenceType(XmlTypeCode.Time, XmlTypeCardinality.ZeroOrOne))));
            Add(XmlReservedNs.NsXQueryFunc, "timezone-from-time", 1, (context, provider, args) =>
                ExtFuncs.TimezoneFromTime(CoreFuncs.CastArg(context, CoreFuncs.Atomize(args[0]),
                    new SequenceType(XmlTypeCode.Time, XmlTypeCardinality.ZeroOrOne))));
            Add(XmlReservedNs.NsXQueryFunc, "adjust-dateTime-to-timezone", 2, (context, provider, args) =>
                ExtFuncs.AdjustDateTimeToTimezone(CoreFuncs.CastArg(context, CoreFuncs.Atomize(args[0]),
                    new SequenceType(XmlTypeCode.DateTime, XmlTypeCardinality.ZeroOrOne)),
                    CoreFuncs.CastArg(context, CoreFuncs.Atomize(args[1]),
                        new SequenceType(XmlTypeCode.DayTimeDuration, XmlTypeCardinality.ZeroOrOne))));
            Add(XmlReservedNs.NsXQueryFunc, "adjust-dateTime-to-timezone", 1, (context, provider, args) => 
                ExtFuncs.AdjustDateTimeToTimezone(CoreFuncs.CastArg(context, CoreFuncs.Atomize(args[0]), 
                    new SequenceType(XmlTypeCode.DateTime, XmlTypeCardinality.ZeroOrOne))));
            Add(XmlReservedNs.NsXQueryFunc, "adjust-date-to-timezone", 2, (context, provider, args) => 
                ExtFuncs.AdjustDateToTimezone(CoreFuncs.CastArg(context, CoreFuncs.Atomize(args[0]),
                    new SequenceType(XmlTypeCode.Date, XmlTypeCardinality.ZeroOrOne)),
                    CoreFuncs.CastArg(context, CoreFuncs.Atomize(args[1]), new SequenceType(XmlTypeCode.DayTimeDuration, 
                         XmlTypeCardinality.ZeroOrOne))));
            Add(XmlReservedNs.NsXQueryFunc, "adjust-date-to-timezone", 1, (context, provider, args) =>
                ExtFuncs.AdjustDateToTimezone(CoreFuncs.CastArg(context, CoreFuncs.Atomize(args[0]),
                    new SequenceType(XmlTypeCode.Date, XmlTypeCardinality.ZeroOrOne))));
            Add(XmlReservedNs.NsXQueryFunc, "adjust-time-to-timezone", 2, (context, provider, args) =>
                ExtFuncs.AdjustTimeToTimezone(CoreFuncs.CastArg(context, CoreFuncs.Atomize(args[0]),
                    new SequenceType(XmlTypeCode.Time, XmlTypeCardinality.ZeroOrOne)),
                    CoreFuncs.CastArg(context, CoreFuncs.Atomize(args[1]), 
                         new SequenceType(XmlTypeCode.DayTimeDuration, XmlTypeCardinality.ZeroOrOne))));
            Add(XmlReservedNs.NsXQueryFunc, "adjust-time-to-timezone", 1, (context, provider, args) =>
                ExtFuncs.AdjustTimeToTimezone(CoreFuncs.CastArg(context, CoreFuncs.Atomize(args[0]),
                    new SequenceType(XmlTypeCode.Time, XmlTypeCardinality.ZeroOrOne))));
            Add(XmlReservedNs.NsXQueryFunc, "abs", 1, (context, provider, args) =>
                ExtFuncs.GetAbs(CoreFuncs.Atomize(args[0])));
            Add(XmlReservedNs.NsXQueryFunc, "ceiling", 1, (context, provider, args) =>
                ExtFuncs.GetCeiling(CoreFuncs.Atomize(args[0])));
            Add(XmlReservedNs.NsXQueryFunc, "floor", 1, (context, provider, args) =>
                ExtFuncs.GetFloor(CoreFuncs.Atomize(args[0])));
            Add(XmlReservedNs.NsXQueryFunc, "round", 1, (context, provider, args) =>
                ExtFuncs.GetRound(CoreFuncs.Atomize(args[0])));
            Add(XmlReservedNs.NsXQueryFunc, "round-half-to-even", 2, (context, provider, args) =>
                ExtFuncs.GetRoundHalfToEven(CoreFuncs.Atomize(args[0]), CoreFuncs.Atomize(args[1])));
            Add(XmlReservedNs.NsXQueryFunc, "round-half-to-even", 1, (context, provider, args) => 
                ExtFuncs.GetRoundHalfToEven(CoreFuncs.Atomize(args[0])));
            Add(XmlReservedNs.NsXQueryFunc, "compare", 2, (context, provider, args) =>
                ExtFuncs.Compare(CoreFuncs.CastString(context, args[0]), CoreFuncs.CastString(context, args[1]))); 
            Add(XmlReservedNs.NsXQueryFunc, "compare", 3, (context, provider, args) =>
                ExtFuncs.Compare(context, CoreFuncs.CastString(context, args[0]), CoreFuncs.CastString(context, args[1]),
                    (String)CoreFuncs.CastArg(context, args[2], new SequenceType(XmlTypeCode.String))));
            Add(XmlReservedNs.NsXQueryFunc, "codepoint-equal", 2, (context, provider, args) =>
                ExtFuncs.CodepointEqual(CoreFuncs.CastString(context, args[0]), CoreFuncs.CastString(context, args[1])));
            Add(XmlReservedNs.NsXQueryFunc, "empty", 1, (context, provider, args) => 
                ExtFuncs.EmptySequence(XPath2NodeIterator.Create(args[0])));
            Add(XmlReservedNs.NsXQueryFunc, "exists", 1, (context, provider, args) =>
                ExtFuncs.ExistsSequence(XPath2NodeIterator.Create(args[0])));
            Add(XmlReservedNs.NsXQueryFunc, "reverse", 1, (context, provider, args) =>
                ExtFuncs.ReverseSequence(XPath2NodeIterator.Create(args[0])));
            Add(XmlReservedNs.NsXQueryFunc, "index-of", 3, (context, provider, args) => 
                ExtFuncs.IndexOfSequence(context, XPath2NodeIterator.Create(args[0]), CoreFuncs.Atomize(args[1]),
                    (String)CoreFuncs.CastArg(context, args[2], new SequenceType(XmlTypeCode.String))));
            Add(XmlReservedNs.NsXQueryFunc, "index-of", 2, (context, provider, args) => 
                ExtFuncs.IndexOfSequence(XPath2NodeIterator.Create(args[0]), CoreFuncs.Atomize(args[1])));
            Add(XmlReservedNs.NsXQueryFunc, "remove", 2, (context, provider, args) =>
                ExtFuncs.Remove(XPath2NodeIterator.Create(args[0]), (int)CoreFuncs.CastArg(context, args[1], SequenceType.Int)));
            Add(XmlReservedNs.NsXQueryFunc, "insert-before", 3, (context, provider, args) =>
                ExtFuncs.InsertBefore(XPath2NodeIterator.Create(args[0]), (int)CoreFuncs.CastArg(context, args[1], SequenceType.Int),
                    XPath2NodeIterator.Create(args[2])));
            Add(XmlReservedNs.NsXQueryFunc, "subsequence", 3, (context, provider, args) =>
                ExtFuncs.Subsequence(XPath2NodeIterator.Create(args[0]), CoreFuncs.Number(context, args[1]),
                    CoreFuncs.Number(context, args[2])));
            Add(XmlReservedNs.NsXQueryFunc, "subsequence", 2, (context, provider, args) =>
                ExtFuncs.Subsequence(XPath2NodeIterator.Create(args[0]), CoreFuncs.Number(context, args[1])));
            Add(XmlReservedNs.NsXQueryFunc, "unordered", 1, (context, provider, args) =>
                ExtFuncs.Unordered(XPath2NodeIterator.Create(args[0])));
            Add(XmlReservedNs.NsXQueryFunc, "zero-or-one", 1, (context, provider, args) =>
                ExtFuncs.ZeroOrOne(XPath2NodeIterator.Create(args[0])));
            Add(XmlReservedNs.NsXQueryFunc, "one-or-more", 1, (context, provider, args) =>
                ExtFuncs.OneOrMore(XPath2NodeIterator.Create(args[0])));
            Add(XmlReservedNs.NsXQueryFunc, "exactly-one", 1, (context, provider, args) =>
                ExtFuncs.ExactlyOne(XPath2NodeIterator.Create(args[0])));
            Add(XmlReservedNs.NsXQueryFunc, "distinct-values", 2, (context, provider, args) =>
                ExtFuncs.DistinctValues(context, XPath2NodeIterator.Create(args[0]), 
                    (String)CoreFuncs.CastArg(context, args[1], new SequenceType(XmlTypeCode.String))));
            Add(XmlReservedNs.NsXQueryFunc, "distinct-values", 1, (context, provider, args) => 
                ExtFuncs.DistinctValues(XPath2NodeIterator.Create(args[0])));
            Add(XmlReservedNs.NsXQueryFunc, "deep-equal", 3, (context, provider, args) => 
                ExtFuncs.DeepEqual(context, XPath2NodeIterator.Create(args[0]), XPath2NodeIterator.Create(args[1]),
                (String)CoreFuncs.CastArg(context, CoreFuncs.Atomize(args[2]), new SequenceType(XmlTypeCode.String))));
            Add(XmlReservedNs.NsXQueryFunc, "deep-equal", 2, (context, provider, args) => 
                ExtFuncs.DeepEqual(XPath2NodeIterator.Create(args[0]), XPath2NodeIterator.Create(args[1])));
            Add(XmlReservedNs.NsXQueryFunc, "count", 1, (context, provider, args) =>
                ExtFuncs.CountValues(XPath2NodeIterator.Create(args[0])));
            Add(XmlReservedNs.NsXQueryFunc, "max", 1, (context, provider, args) => 
                ExtFuncs.MaxValue(context, XPath2NodeIterator.Create(args[0])));
            Add(XmlReservedNs.NsXQueryFunc, "max", 2, (context, provider, args) => 
                ExtFuncs.MaxValue(context, XPath2NodeIterator.Create(args[0]),
                (String)CoreFuncs.CastArg(context, CoreFuncs.Atomize(args[1]), new SequenceType(XmlTypeCode.String))));
            Add(XmlReservedNs.NsXQueryFunc, "min", 1, (context, provider, args) =>
                ExtFuncs.MinValue(context, XPath2NodeIterator.Create(args[0])));
            Add(XmlReservedNs.NsXQueryFunc, "min", 2, (context, provider, args) =>
                ExtFuncs.MaxValue(context, XPath2NodeIterator.Create(args[0]),
                (String)CoreFuncs.CastArg(context, CoreFuncs.Atomize(args[1]), new SequenceType(XmlTypeCode.String))));
            Add(XmlReservedNs.NsXQueryFunc, "sum", 2, (context, provider, args) => 
                ExtFuncs.SumValue(context, XPath2NodeIterator.Create(args[0]), CoreFuncs.Atomize(args[1])));
            Add(XmlReservedNs.NsXQueryFunc, "sum", 1, (context, provider, args) => 
                ExtFuncs.SumValue(context, XPath2NodeIterator.Create(args[0])));
            Add(XmlReservedNs.NsXQueryFunc, "avg", 1, (context, provider, args) =>
                ExtFuncs.AvgValue(context, XPath2NodeIterator.Create(args[0])));
            Add(XmlReservedNs.NsXQueryFunc, "position", 0, (context, provider, args) =>
                ExtFuncs.CurrentPosition(provider));
            Add(XmlReservedNs.NsXQueryFunc, "last", 0, (context, provider, args) =>
                ExtFuncs.LastPosition(provider));
            Add(XmlReservedNs.NsXQueryFunc, "root", 1, (context, provider, args) =>
                CoreFuncs.GetRoot(CoreFuncs.NodeValue(args[0], false)));
            Add(XmlReservedNs.NsXQueryFunc, "boolean", 1, (context, provider, args) =>
                CoreFuncs.BooleanValue(args[0]));
            Add(XmlReservedNs.NsXQueryFunc, "true", 0, (context, provider, args) => CoreFuncs.True);
            Add(XmlReservedNs.NsXQueryFunc, "false", 0, (context, provider, args) => CoreFuncs.False);
            Add(XmlReservedNs.NsXQueryFunc, "not", 1, (context, provider, args) => 
                CoreFuncs.Not(args[0]));
            Add(XmlReservedNs.NsXQueryFunc, "string", 1, (context, provider, args) =>
                CoreFuncs.StringValue(context, args[0]));
            Add(XmlReservedNs.NsXQueryFunc, "number", 0, (context, provider, args) =>
                CoreFuncs.Number(context, provider));
            Add(XmlReservedNs.NsXQueryFunc, "number", 1, (context, provider, args) =>
                CoreFuncs.Number(context, args[0]));
        }        

        internal XPathFunctionDelegate Bind(string name, string ns, int arity)
        {
            XPathFunctionDelegate res;
            if (!_funcTable.TryGetValue(new FunctionDesc(name, ns, arity), out res))
                return null;
            return res;
        }

        internal void Add(string ns, string name, XPathFunctionDelegate action)
        {
            Add(ns, name, -1, action);
        }

        internal void Add(string ns, string name, int arity, XPathFunctionDelegate action)
        {
            _funcTable.Add(new FunctionDesc(name, ns, arity), action);
        }

        private static object syncRoot = new Object();
        private static volatile FunctionTable _inst = null;

        public static FunctionTable Inst
        {
            get
            {
                if (_inst == null)
                {
                    lock (syncRoot)
                        if (_inst == null)
                            _inst = new FunctionTable();
                }
                return _inst;
            }
        }

        private sealed class FunctionDesc
        {
            public readonly String name;
            public readonly String ns;
            public readonly int arity;

            public FunctionDesc(String name, String ns, int arity)
            {
                this.name = name;
                this.ns = ns;
                this.arity = arity;
            }

            public override bool Equals(object obj)
            {
                FunctionDesc other = obj as FunctionDesc;
                if (other != null)
                    return name == other.name && ns == other.ns
                        && (arity == other.arity || arity == -1 || other.arity == -1);
                return false;
            }

            public override int GetHashCode()
            {
                return name.GetHashCode();
            }
        }
    }
}
